@@ -5,9 +5,7 @@
  *
  *	Created by:	John T. Kohl
  *
- *	$Source: /mit/zephyr/repository/zephyr/clients/zstat/zserver.h,v $
- *	$Author: ghudson $
- *	$Header: /mit/zephyr/repository/zephyr/clients/zstat/zserver.h,v 1.32 1995/06/30 21:50:07 ghudson Exp $
+ *	$Id: zserver.h 2091 2007-12-20 01:17:23Z kcr $
  *
  *	Copyright (c) 1987,1988 by the Massachusetts Institute of Technology.
  *	For copying and distribution information, see the file
