@@ -6,9 +6,7 @@
  *
  * Created by Ken Raeburn.
  *
- * $Source: /mit/zephyr/repository/zephyr/server/access.h,v $
- * $Author: ghudson $
- * $Id: access.h,v 1.4 1995/06/30 22:10:55 ghudson Exp $
+ * $Id: access.h 2527 2009-08-09 18:42:32Z kcr@ATHENA.MIT.EDU $
  *
  * Copyright (c) 1990 by the Massachusetts Institute of Technology.
  * For copying and distribution information, see the file
@@ -33,11 +31,11 @@ typedef struct _Acl {
 } Acl;
 
 /* found in access.c */
-void access_init __P((void));
-void access_reinit __P((void));
+void access_init(void);
+void access_reinit(void);
 
 /* found in acl_files.c */
-int acl_load __P((char *));
+int acl_load(char *);
 
 /* external data relevant */
 extern int zdebug;
