@@ -3,12 +3,12 @@
  *
  *	Created by:	John T. Kohl
  *
- *	$Source: /mit/zephyr/repository/zephyr/server/server.c,v $
- *	$Author: ghudson $
+ *	$Source: /afs/dev.mit.edu/source/repository/athena/lib/zephyr/server/server.c,v $
+ *	$Author: kcr@ATHENA.MIT.EDU $
  *
  *	Copyright (c) 1987, 1991 by the Massachusetts Institute of Technology.
  *	For copying and distribution information, see the file
- *	"mit-copyright.h". 
+ *	"mit-copyright.h".
  */
 
 #include <zephyr/mit-copyright.h>
@@ -17,10 +17,20 @@
 
 #ifndef lint
 #ifndef SABER
-static const char rcsid_server_c[] = "$Id: server.c,v 1.64 1995/07/07 22:12:18 ghudson Exp $";
+static const char rcsid_server_c[] = "$Id: server.c 2527 2009-08-09 18:42:32Z kcr@ATHENA.MIT.EDU $";
 #endif
 #endif
 
+enum {
+    SRV_NACKTAB_HASHSIZE = 1023
+};
+inline unsigned int
+srv_nacktab_hashval(int which, ZUnique_Id_t uid) {
+    return (which ^
+	    uid.zuid_addr.s_addr ^ uid.tv.tv_sec ^ uid.tv.tv_usec)
+	% SRV_NACKTAB_HASHSIZE;
+}
+
 /*
  * Server manager.  Deal with  traffic to and from other servers.
  *
@@ -62,39 +72,33 @@ static const char rcsid_server_c[] = "$Id: server.c,v 1.64 1995/07/07 22:12:18 g
  * void server_reset();
  */
 
-static void server_flush __P((Server *));
-static void hello_respond __P((struct sockaddr_in *, int, int));
-static void srv_responded __P((struct sockaddr_in *));
-static void send_msg __P((struct sockaddr_in *, char *, int));
-static void send_msg_list __P((struct sockaddr_in *, char *, char **, int,
-			       int, int));
-static void srv_nack_cancel __P((ZNotice_t *, struct sockaddr_in *));
-static void srv_nack_release __P((Server *));
-static void srv_nack_renumber  __P((int *));
-static void send_stats __P((struct sockaddr_in *));
-static void server_queue __P((Server *, int, void *, int,
-			      struct sockaddr_in *));
-static void server_hello __P((Server *, int));
-static void setup_server __P((Server *, struct in_addr *));
-static void srv_rexmit __P((void *));
-static void server_forw_reliable __P((Server *, caddr_t, int, ZNotice_t *));
-static Code_t admin_dispatch __P((ZNotice_t *, int, struct sockaddr_in *,
-				  Server *));
-static Code_t kill_clt __P((ZNotice_t *, Server *));
-static Code_t extract_addr __P((ZNotice_t *, struct sockaddr_in *));
+static void server_flush(Server *);
+static void hello_respond(struct sockaddr_in *, int, int);
+static void srv_responded(struct sockaddr_in *);
+static void send_msg(struct sockaddr_in *, char *, int);
+static void send_msg_list(struct sockaddr_in *, char *, char **, int,
+			       int);
+static void srv_nack_cancel(ZNotice_t *, struct sockaddr_in *);
+static void srv_nack_release(Server *);
+static void srv_nack_renumber (int *);
+static void send_stats(struct sockaddr_in *);
+static void server_queue(Server *, int, void *, int,
+			      struct sockaddr_in *);
+static void server_hello(Server *, int);
+static void setup_server(Server *, struct in_addr *);
+static void srv_rexmit(void *);
+static void server_forw_reliable(Server *, caddr_t, int, ZNotice_t *);
+static Code_t admin_dispatch(ZNotice_t *, int, struct sockaddr_in *,
+				  Server *);
+static Code_t kill_clt(ZNotice_t *, Server *);
+static Code_t extract_addr(ZNotice_t *, struct sockaddr_in *);
 
-#ifdef notdef
-static Code_t server_register();
-#endif
+static struct in_addr *get_server_addrs(int *number);
+static char **get_server_list(char *file);
+static char **get_single_server(void);
+static void free_server_list(char **list);
 
-static struct in_addr *get_server_addrs __P((int *number));
-#ifndef ZEPHYR_USES_HESIOD
-static char **get_server_list __P((char *file));
-static void free_server_list __P((char **list));
-#endif
-
-Unacked *srv_nacklist = NULL;	/* not acked list for server-server
-				   packets */
+static Unacked *srv_nacktab[SRV_NACKTAB_HASHSIZE];
 Server *otherservers;		/* points to an array of the known
 				   servers */
 int nservers;			/* number of other servers */
@@ -131,7 +135,7 @@ int zalone;
  */
 
 void
-server_init()
+server_init(void)
 {
     int i;
     struct in_addr *serv_addr, *server_addrs, limbo_addr;
@@ -175,9 +179,6 @@ server_init()
 	    otherservers[i].timer = NULL;
 	    otherservers[i].queue = NULL;
 	    otherservers[i].dumping = 0;
-#if 0
-	    zdbug((LOG_DEBUG,"found myself"));
-#endif
 	}
     }
 
@@ -227,7 +228,7 @@ server_init()
  * handle on a particular server other than by indexing on otherservers[].
  */
 void
-server_reset()
+server_reset(void)
 {
     int num_servers;
     struct in_addr *server_addrs;
@@ -237,9 +238,6 @@ server_reset()
     int *ok_list_new, *ok_list_old;
     int num_ok, new_num;
 
-#if 0
-    zdbug((LOG_DEBUG, "server_reset"));
-#endif
 #ifdef DEBUG
     if (zalone) {
 	syslog(LOG_INFO, "server_reset while alone, punt");
@@ -267,7 +265,7 @@ server_reset()
 
     memset(ok_list_old, 0, nservers * sizeof(int));
     memset(ok_list_new, 0, num_servers * sizeof(int));
-	
+
     /* reset timers--pointers will move */
     for (j = 1; j < nservers; j++) {	/* skip limbo */
 	if (j == me_server_idx)
@@ -395,18 +393,11 @@ server_reset()
 	if (i != me_server_idx && !otherservers[i].timer) {
 	    otherservers[i].timer =
 		timer_set_rel(0L, server_timo, &otherservers[i]);
-#if 0
-	    zdbug((LOG_DEBUG, "reset timer for %s",
-		   otherservers[i].addr_str));
-#endif	
 	}
     }
     free(ok_list_old);
     free(ok_list_new);
 
-#if 0
-    zdbug((LOG_DEBUG, "server_reset: %d servers now", nservers));
-#endif
 }
 
 /* note: these must match the order given in zserver.h */
@@ -417,26 +408,29 @@ srv_states[] = {
     "SERV_DEAD",
     "SERV_STARTING"
 };
+static char *
+rlm_states[] = {
+    "REALM_UP",
+    "REALM_TARDY",
+    "REALM_DEAD",
+    "REALM_STARTING"
+};
 
-/* 
+/*
  * A server timout has expired.  If enough hello's have been unanswered,
  * change state and act accordingly. Send a "hello" and reset the timer,
  * incrementing the number of hello's sent.
  *
  * See the FSM in the Zephyr document for a better picture of what's
- * happening here. 
+ * happening here.
  */
 
 void
-server_timo(arg)
-    void *arg;
+server_timo(void *arg)
 {
     Server *which = (Server *) arg;
     int auth = 0;
 
-#if 0
-    zdbug((LOG_DEBUG,"srv_timo: %s", which->addr_str));
-#endif
     /* change state and reset if appropriate */
     switch(which->state) {
       case SERV_DEAD:			/* leave him dead */
@@ -463,15 +457,11 @@ server_timo(arg)
 	auth = 0;
 	break;
       default:
-	syslog(LOG_ERR,"Bad server state, server 0x%x\n",which);
+	syslog(LOG_ERR,"Bad server state, server 0x%x\n", (int)which);
 	abort();
     }
     /* now he's either TARDY, STARTING, or DEAD
        We send a "hello," which increments the counter */
-#if 0
-    zdbug((LOG_DEBUG, "srv %s is %s", which->addr_str,
-	   srv_states[which->state]));
-#endif
     server_hello(which, auth);
     /* reschedule the timer */
     which->timer = timer_set_rel(which->timeout, server_timo, which);
@@ -483,19 +473,15 @@ server_timo(arg)
 
 /*ARGSUSED*/
 Code_t
-server_dispatch(notice, auth, who)
-    ZNotice_t *notice;
-    int auth;
-    struct sockaddr_in *who;
+server_dispatch(ZNotice_t *notice,
+		int auth,
+		struct sockaddr_in *who)
 {
     Server *server;
     struct sockaddr_in newwho;
     Code_t status;
     String *notice_class;
 
-#if 0
-    zdbug((LOG_DEBUG, "server_dispatch"));
-#endif
 
     if (notice->z_kind == SERVACK) {
 	srv_nack_cancel(notice, who);
@@ -503,10 +489,7 @@ server_dispatch(notice, auth, who)
 	return ZERR_NONE;
     }
     /* set up a who for the real origin */
-    memset(&newwho, 0, sizeof(newwho));
-    newwho.sin_family = AF_INET;
-    newwho.sin_addr.s_addr = notice->z_sender_addr.s_addr;
-    newwho.sin_port = notice->z_port;
+    notice_extract_address(notice, &newwho);
 
     server = server_which_server(who);
 
@@ -517,7 +500,9 @@ server_dispatch(notice, auth, who)
 
     notice_class = make_string(notice->z_class, 1);
 
-    if (class_is_admin(notice_class)) {
+    if (realm_which_realm(&newwho))
+	status = realm_dispatch(notice, auth, &newwho, server);
+    else if (class_is_admin(notice_class)) {
 	/* admins don't get acked, else we get a packet loop */
 	/* will return  requeue if bdump request and dumping */
 	i_s_admins.val++;
@@ -542,80 +527,12 @@ server_dispatch(notice, auth, who)
     return status;
 }
 
-#ifdef notdef
-/*
- * Register a new server (one not in our list).  This MUST be authenticated.
- */
-
-/*ARGSUSED*/
-static Code_t
-server_register(notice, auth, who)
-    ZNotice_t *notice;
-    int auth;
-    struct sockaddr_in *who;
-{
-    Server *temp;
-    int i;
-    long timerval;
-
-    if (who->sin_port != srv_addr.sin_port) {
-#if 0
-	zdbug((LOG_DEBUG, "srv_wrong port %d", ntohs(who->sin_port)));
-#endif
-	return 1;
-    }
-    /* Not yet... talk to ken about authenticators */
-#ifdef notdef
-    if (!auth) {
-#if 0
-	zdbug((LOG_DEBUG, "srv_unauth"));
-#endif
-	return 1;
-    }
-#endif /* notdef */
-    /* OK, go ahead and set him up. */
-    temp = (Server *) malloc((nservers + 1) * sizeof(Server));
-    if (!temp) {
-	syslog(LOG_CRIT, "srv_reg malloc");
-	return 1;
-    }
-
-    memcpy(temp, otherservers, nservers * sizeof(Server));
-    free(otherservers);
-    otherservers = temp;
-    /* don't reschedule limbo's timer, so start i=1 */
-    for (i = 1; i < nservers; i++) {
-	if (i == me_server_idx) /* don't reset myself */
-	    continue;
-	/* reschedule the timers--we moved otherservers */
-	timerval = timer_when(otherservers[i].timer);
-	timer_reset(otherservers[i].timer);
-	otherservers[i].timer = timer_set_abs(timerval, server_timo,
-					      &otherservers[i]);
-    }
-    setup_server(&otherservers[nservers], &who->sin_addr);
-    otherservers[nservers].state = SERV_STARTING;
-    otherservers[nservers].timeout = timo_tardy;
-    otherservers[nservers].update_queue = NULL;
-    otherservers[nservers].dumping = 0;
-
-    nservers++;
-#if 0
-    zdbug((LOG_DEBUG, "srv %s is %s", otherservers[nservers].addr_str,
-	   srv_states[otherservers[nservers].state]));
-#endif
-
-    return 0;
-}
-#endif
-
 /*
  * Tell the other servers that this client died.
  */
 
 void
-server_kill_clt(client)
-    Client *client;
+server_kill_clt(Client *client)
 {
     int i;
     char buf[512], *lyst[2];
@@ -629,12 +546,10 @@ server_kill_clt(client)
     sprintf(buf, "%d", ntohs(client->addr.sin_port));
     lyst[1] = buf;
 
-#if 0
-    zdbug((LOG_DEBUG, "server kill clt %s/%s", lyst[0], lyst[1]));
-#endif
-
     pnotice = &notice;
 
+    memset (&notice, 0, sizeof(notice));
+
     pnotice->z_kind = ACKED;
 
     pnotice->z_port = srv_addr.sin_port;
@@ -671,31 +586,26 @@ server_kill_clt(client)
  */
 
 static Code_t
-kill_clt(notice, server)
-    ZNotice_t *notice;
-    Server *server;
+kill_clt(ZNotice_t *notice,
+	 Server *server)
 {
     struct sockaddr_in who;
     Client *client;
 
-#if 0
-    zdbug((LOG_DEBUG, "kill_clt"));
-#endif
     if (extract_addr(notice, &who) != ZERR_NONE)
 	return ZERR_NONE;	/* XXX */
-    client = client_which_client(&who.sin_addr, notice);
+    client = client_find(&who.sin_addr, notice->z_port);
     if (!client) {
 	syslog(LOG_NOTICE, "kill_clt: no such client (%s/%d) from %s",
 	       inet_ntoa(who.sin_addr), ntohs(who.sin_port),
 	       server->addr_str);
 	return ZERR_NONE;	/* XXX */
     }
-#if 1
-    if (zdebug || 1) {
+
+    if (zdebug) {
 	syslog(LOG_DEBUG, "kill_clt clt_dereg %s/%d from %s",
 	       inet_ntoa(who.sin_addr), ntohs(who.sin_port), server->addr_str);
     }
-#endif
 
     /* remove the locations, too */
     client_deregister(client, 1);
@@ -707,9 +617,8 @@ kill_clt(notice, server)
  */
 
 static Code_t
-extract_addr(notice, who)
-    ZNotice_t *notice;
-    struct sockaddr_in *who;
+extract_addr(ZNotice_t *notice,
+	     struct sockaddr_in *who)
 {
     char *cp = notice->z_message;
 
@@ -726,10 +635,6 @@ extract_addr(notice, who)
     }
     who->sin_port = notice->z_port = htons((u_short) atoi(cp));
     who->sin_family = AF_INET;
-#if 0
-    zdbug((LOG_DEBUG,"ext %s/%d", inet_ntoa(who->sin_addr),
-	   ntohs(who->sin_port)));
-#endif
     return ZERR_NONE;
 }
 
@@ -738,13 +643,8 @@ extract_addr(notice, who)
  */
 
 static void
-server_flush(which)
-    Server *which;
+server_flush(Server *which)
 {
-#if 0
-    if (zdebug)
-	syslog(LOG_DEBUG, "server_flush %s", which->addr_str);
-#endif
     srv_nack_release(which);
 }
 
@@ -754,9 +654,8 @@ server_flush(which)
  */
 
 static void
-server_hello(which, auth)
-    Server *which;
-    int auth;
+server_hello(Server *which,
+	     int auth)
 {
     send_msg(&which->addr, ADMIN_HELLO, auth);
     which->num_hello_sent++;
@@ -768,40 +667,31 @@ server_hello(which, auth)
 
 /*ARGSUSED*/
 static Code_t
-admin_dispatch(notice, auth, who, server)
-    ZNotice_t *notice;
-    int auth;
-    struct sockaddr_in *who;
-    Server *server;
+admin_dispatch(ZNotice_t *notice,
+	       int auth,
+	       struct sockaddr_in *who,
+	       Server *server)
 {
     char *opcode = notice->z_opcode;
     Code_t status = ZERR_NONE;
 
-#if 0
-    zdbug((LOG_DEBUG, "ADMIN received"));
-#endif
-
     if (strcmp(opcode, ADMIN_HELLO) == 0) {
 	hello_respond(who, ADJUST, auth);
     } else if (strcmp(opcode, ADMIN_IMHERE) == 0) {
 	srv_responded(who);
     } else if (strcmp(opcode, ADMIN_SHUTDOWN) == 0) {
-#if 0
-	zdbug((LOG_DEBUG, "server shutdown"));
-#endif
 	if (server) {
 	    srv_nack_release(server);
 	    server->state = SERV_DEAD;
 	    server->timeout = timo_dead;
 	    /* don't worry about the timer, it will
 	       be set appropriately on the next send */
-#if 0
-	    zdbug((LOG_DEBUG, "srv %s is %s", server->addr_str,
-		   srv_states[server->state]));
-#endif
-		}
+	}
     } else if (strcmp(opcode, ADMIN_BDUMP) == 0) {
-	if (bdumping)
+	/* Ignore a brain dump request if this is a brain dump packet
+         * or a packet being processed concurrently during a brain
+         * dump. */
+	if (bdumping || bdump_concurrent)
 	    return ZERR_NONE;
 	bdump_get(notice, auth, who, server);
     } else if (strcmp(opcode, ADMIN_KILL_CLT) == 0) {
@@ -823,11 +713,10 @@ admin_dispatch(notice, auth, who, server)
 
 /*ARGSUSED*/
 Code_t
-server_adispatch(notice, auth, who, server)
-    ZNotice_t *notice;
-    int auth;
-    struct sockaddr_in *who;
-    Server *server;
+server_adispatch(ZNotice_t *notice,
+		 int auth,
+		 struct sockaddr_in *who,
+		 Server *server)
 {
 
     /* this had better be a HELLO message--start of acquisition
@@ -839,45 +728,32 @@ server_adispatch(notice, auth, who, server)
 	return ZERR_NONE;
     }
 
-#ifdef notdef
-    syslog(LOG_INFO, "disp: new server?");
-    if (server_register(notice, auth, who) != ZERR_NONE) {
-	syslog(LOG_INFO, "new server failed");
-    } else {
-	syslog(LOG_INFO, "new server %s, %d", inet_ntoa(who->sin_addr),
-	       ntohs(who->sin_port));
-	hello_respond(who, DONT_ADJUST, auth);
-    }
-#else
     syslog(LOG_INFO, "srv_adisp: server attempt from %s",
 	   inet_ntoa(who->sin_addr));
-#endif /* notdef */
 
     return ZERR_NONE;
 }
 
 static void
-send_stats(who)
-    struct sockaddr_in *who;
+send_stats(struct sockaddr_in *who)
 {
     int i;
     char buf[BUFSIZ];
     char **responses;
     int num_resp;
     char *vers, *pkts, *upt;
+    ZRealm *realm;
 
-#if defined(OLD_COMPAT) || defined(NEW_COMPAT)
     int extrafields = 0;
-#endif /* OLD_ or NEW_COMPAT */
 #define	NUM_FIXED 3			/* 3 fixed fields, plus server info */
 					/* well, not really...but for
 					   backward compatibility, we gotta
 					   do it this way. */
     vers = get_version();
 
-    sprintf(buf, "%d pkts", npackets);
+    sprintf(buf, "%lu pkts", npackets);
     pkts = strsave(buf);
-    sprintf(buf, "%d seconds operational",NOW - uptime);
+    sprintf(buf, "%ld seconds operational",NOW - uptime);
     upt = strsave(buf);
 
 #ifdef OLD_COMPAT
@@ -894,12 +770,9 @@ send_stats(who)
     if (new_compat_count_subscr)
 	extrafields++;
 #endif /* NEW_COMPAT */
-#if defined(OLD_COMPAT) || defined(NEW_COMPAT)
+    extrafields += nrealms;
     responses = (char **) malloc((NUM_FIXED + nservers + extrafields) *
 				 sizeof(char *));
-#else
-    responses = (char **) malloc((NUM_FIXED + nservers) * sizeof(char *));
-#endif /* OLD_ or NEW_COMPAT */
     responses[0] = vers;
     responses[1] = pkts;
     responses[2] = upt;
@@ -937,8 +810,14 @@ send_stats(who)
 	responses[num_resp++] = strsave(buf);
     }
 #endif /* NEW_COMPAT */
+    for (realm = otherrealms, i = 0; i < nrealms ; i++, realm++) {
+      sprintf(buf, "%s(%s)/%s", realm->name,
+	      inet_ntoa((realm->addrs[realm->idx]).sin_addr),
+	      rlm_states[(int) realm->state]);
+      responses[num_resp++] = strsave(buf);
+    }
 
-    send_msg_list(who, ADMIN_STATUS, responses, num_resp, 0, -1);
+    send_msg_list(who, ADMIN_STATUS, responses, num_resp, 0);
 
     /* Start at one; don't try to free static version string */
     for (i = 1; i < num_resp; i++)
@@ -948,7 +827,7 @@ send_stats(who)
 
 /*
  * Get a list of server addresses.
-#ifdef ZEPHYR_USES_HESIOD
+#ifdef HAVE_HESIOD
  * This list is retrieved from Hesiod.
 #else
  * This list is read from a file.
@@ -958,31 +837,33 @@ send_stats(who)
  */
 
 static struct in_addr *
-get_server_addrs(number)
-    int *number; /* RETURN */
+get_server_addrs(int *number)
 {
     int i;
-    char **server_hosts;
+    char **server_hosts = NULL;
+    char **server_hosts_free = NULL;
     char **cpp;
     struct in_addr *addrs;
     struct in_addr *addr;
     struct hostent *hp;
 
-#ifdef ZEPHYR_USES_HESIOD
-    /* get the names from Hesiod */
-    server_hosts = hes_resolve("zephyr","sloc");
-    if (!server_hosts)
-	return NULL;
-#else
     server_hosts = get_server_list(list_file);
+    server_hosts_free = server_hosts;
+#ifdef HAVE_HESIOD
+    if (!server_hosts)
+	server_hosts = hes_resolve("zephyr","sloc");
+#endif
+    if (!server_hosts) {
+	server_hosts = get_single_server();
+	server_hosts_free = server_hosts;
+    }
     if (!server_hosts)
 	return NULL;
-#endif
     /* count up */
     i = 0;
     for (cpp = server_hosts; *cpp; cpp++)
 	i++;
-	
+
     addrs = (struct in_addr *) malloc(i * sizeof(struct in_addr));
 
     /* Convert to in_addr's */
@@ -996,14 +877,11 @@ get_server_addrs(number)
 	}
     }
     *number = i;
-#ifndef ZEPHYR_USES_HESIOD
-    free_server_list(server_hosts);
-#endif
+    if (server_hosts_free)
+      free_server_list(server_hosts_free);
     return addrs;
 }
 
-#ifndef ZEPHYR_USES_HESIOD
-
 static int nhosts = 0;
 
 /*
@@ -1012,8 +890,7 @@ static int nhosts = 0;
  */
 
 static char **
-get_server_list(file)
-    char *file;
+get_server_list(char *file)
 {
     FILE *fp;
     char buf[MAXHOSTNAMELEN];
@@ -1023,11 +900,12 @@ get_server_list(file)
 
     fp = fopen(file, "r");
     if (!fp)
-	return NULL;
-
+    	return NULL;
     /* start with 16, realloc if necessary */
     nhosts = 16;
     ret_list = (char **) malloc(nhosts * sizeof(char *));
+    if (!ret_list)
+    	return NULL;
 
     while (fgets(buf, MAXHOSTNAMELEN, fp)) {
 	/* nuke the newline, being careful not to overrun
@@ -1046,16 +924,38 @@ get_server_list(file)
 	ret_list[nused++] = strsave(buf);
     }
     fclose(fp);
+    if (!nused) {
+    	free(ret_list);
+	return NULL;
+    }
     ret_list[nused] = NULL;
     return ret_list;
 }
 
-/* 
+static char **
+get_single_server(void)
+{
+    char buf[MAXHOSTNAMELEN];
+    char **ret_list;
+    int nused = 0;
+    nhosts = 2;
+    ret_list = (char **) malloc(nhosts * sizeof(char *));
+    if (!ret_list)
+    	return NULL;
+    if (gethostname(buf, sizeof(buf)) < 0) {
+	free(ret_list);
+	return NULL;
+    }
+    ret_list[nused++] = strsave(buf);
+    ret_list[nused] = NULL;
+    return ret_list;
+}
+
+/*
  * free storage allocated by get_server_list
  */
 static void
-free_server_list(list)
-    char **list;
+free_server_list(char **list)
 {
     char **orig_list = list;
 
@@ -1066,7 +966,6 @@ free_server_list(list)
     free(orig_list);
     return;
 }
-#endif
 
 /*
  * initialize the server structure for address addr, and set a timer
@@ -1074,9 +973,8 @@ free_server_list(list)
  */
 
 static void
-setup_server(server, addr)
-    Server *server;
-    struct in_addr *addr;
+setup_server(Server *server,
+	     struct in_addr *addr)
 {
     server->state = SERV_DEAD;
     server->timeout = timo_dead;
@@ -1096,17 +994,12 @@ setup_server(server, addr)
  */
 
 static void
-hello_respond(who, adj, auth)
-    struct sockaddr_in *who;
-    int adj;
-    int auth;
+hello_respond(struct sockaddr_in *who,
+	      int adj,
+	      int auth)
 {
     Server *which;
 
-#if 0
-    zdbug((LOG_DEBUG, "hello from %s", inet_ntoa(who->sin_addr)));
-#endif
-
     send_msg(who, ADMIN_IMHERE, auth);
     if (adj != ADJUST)
 	return;
@@ -1130,15 +1023,14 @@ hello_respond(who, adj, auth)
       default:
 	break;
     }
-}    
+}
 
 /*
  * return the server descriptor for server at who
  */
 
 Server *
-server_which_server(who)
-    struct sockaddr_in *who;
+server_which_server(struct sockaddr_in *who)
 {
     Server *server;
     int i;
@@ -1159,15 +1051,10 @@ server_which_server(who)
  * appropriately.
  */
 static void
-srv_responded(who)
-    struct sockaddr_in *who;
+srv_responded(struct sockaddr_in *who)
 {
     Server *which = server_which_server(who);
 
-#if 0
-    zdbug((LOG_DEBUG, "srv_responded %s", inet_ntoa(who->sin_addr)));
-#endif
-
     if (!which) {
 	syslog(LOG_ERR, "hello input from non-server?!");
 	return;
@@ -1200,10 +1087,6 @@ srv_responded(who)
 	which->timer = timer_set_rel(which->timeout, server_timo, which);
 	break;
     }
-#if 0
-    zdbug((LOG_DEBUG, "srv %s is %s", which->addr_str,
-	   srv_states[which->state]));
-#endif
 }
 
 /*
@@ -1211,7 +1094,7 @@ srv_responded(who)
  */
 
 void
-server_shutdown()
+server_shutdown(void)
 {
     int i;
 
@@ -1226,10 +1109,9 @@ server_shutdown()
  */
 
 static void
-send_msg(who, opcode, auth)
-    struct sockaddr_in *who;
-    char *opcode;
-    int auth;
+send_msg(struct sockaddr_in *who,
+	 char *opcode,
+	 int auth)
 {
     ZNotice_t notice;
     ZNotice_t *pnotice; /* speed hack */
@@ -1239,6 +1121,8 @@ send_msg(who, opcode, auth)
 
     pnotice = &notice;
 
+    memset (&notice, 0, sizeof(notice));
+
     pnotice->z_kind = ACKED;
 
     pnotice->z_port = srv_addr.sin_port;
@@ -1282,39 +1166,35 @@ send_msg(who, opcode, auth)
  */
 
 static void
-send_msg_list(who, opcode, lyst, num, auth, server_idx)
-    struct sockaddr_in *who;
-    char *opcode;
-    char **lyst;
-    int num;
-    int auth;
-    int server_idx;
+send_msg_list(struct sockaddr_in *who,
+	      char *opcode,
+	      char **lyst,
+	      int num,
+	      int auth)
 {
     ZNotice_t notice;
-    ZNotice_t *pnotice; /* speed hack */
     char *pack;
     int packlen;
     Code_t retval;
-    Unacked *nacked;
 
-    pnotice = &notice;
+    memset (&notice, 0, sizeof(notice));
 
-    pnotice->z_kind = UNSAFE;
-    pnotice->z_port = srv_addr.sin_port;
-    pnotice->z_class = ZEPHYR_ADMIN_CLASS;
-    pnotice->z_class_inst = "";
-    pnotice->z_opcode = opcode;
-    pnotice->z_sender = myname;	/* myname is the hostname */
-    pnotice->z_recipient = "";
-    pnotice->z_default_format = "";
-    pnotice->z_message = NULL;
-    pnotice->z_message_len = 0;
-    pnotice->z_num_other_fields = 0;
+    notice.z_kind = UNSAFE;
+    notice.z_port = srv_addr.sin_port;
+    notice.z_class = ZEPHYR_ADMIN_CLASS;
+    notice.z_class_inst = "";
+    notice.z_opcode = opcode;
+    notice.z_sender = myname;	/* myname is the hostname */
+    notice.z_recipient = "";
+    notice.z_default_format = "";
+    notice.z_message = NULL;
+    notice.z_message_len = 0;
+    notice.z_num_other_fields = 0;
 
     /* XXX for now, we don't do authentication */
     auth = 0;
 
-    retval = ZFormatNoticeList(pnotice, lyst, num, &pack, &packlen,
+    retval = ZFormatNoticeList(&notice, lyst, num, &pack, &packlen,
 			       auth ? ZAUTH : ZNOAUTH);
     if (retval != ZERR_NONE) {
 	syslog(LOG_WARNING, "snd_msg_lst format: %s", error_message(retval));
@@ -1326,36 +1206,8 @@ send_msg_list(who, opcode, lyst, num, auth, server_idx)
 	free(pack);
 	return;
     }
-    retval = ZSendPacket(pack, packlen, 0);
-    if (retval != ZERR_NONE) {
-	syslog(LOG_WARNING, "snd_msg_lst xmit: %s", error_message(retval));
-	free(pack);
-	return;
-    }
-
-    nacked = (Unacked *) malloc(sizeof(Unacked));
-    if (!nacked) {
-	/* no space: just punt */
-	syslog(LOG_WARNING, "xmit nack malloc");
-	free(pack);
-	return;
-    }
-
-    nacked->rexmits = 0;
-    nacked->packet = pack;
-    nacked->packsz = packlen;
-    nacked->uid = pnotice->z_uid;
-
-    /* Set the address and chain into the appropriate queue. */
-    if (server_idx < 0) {
-	nacked->dest.addr = *who;
-	nacked->timer = timer_set_rel(rexmit_times[0], rexmit, nacked);
-	LIST_INSERT(nacklist, nacked);
-    } else {
-	nacked->dest.srv_idx = server_idx;
-	nacked->timer = timer_set_rel(rexmit_times[0], srv_rexmit, nacked);
-	LIST_INSERT(srv_nacklist, nacked);
-    }
+    xmit_frag(&notice, pack, packlen, 0);
+    free(pack);
 }
 
 /*
@@ -1363,19 +1215,15 @@ send_msg_list(who, opcode, lyst, num, auth, server_idx)
  */
 /*ARGSUSED*/
 void
-server_forward(notice, auth, who)
-    ZNotice_t *notice;
-    int auth;
-    struct sockaddr_in *who;
+server_forward(ZNotice_t *notice,
+	       int auth,
+	       struct sockaddr_in *who)
 {
     int i;
     caddr_t pack;
     int packlen;
     Code_t retval;
 
-#if 0
-    zdbug((LOG_DEBUG, "srv_forw"));
-#endif
     /* don't send to limbo */
     for (i = 1; i < nservers; i++) {
 	if (i == me_server_idx)	/* don't xmit to myself */
@@ -1392,7 +1240,7 @@ server_forward(notice, auth, who)
 	    syslog(LOG_CRIT, "srv_fwd malloc");
 	    abort();
 	}
-	retval = ZFormatSmallRawNotice(notice, pack, &packlen);
+	retval = ZNewFormatSmallRawNotice(notice, pack, &packlen);
 	if (retval != ZERR_NONE) {
 	    syslog(LOG_WARNING, "srv_fwd format: %s", error_message(retval));
 	    continue;
@@ -1406,14 +1254,14 @@ server_forward(notice, auth, who)
 }
 
 static void
-server_forw_reliable(server, pack, packlen, notice)
-    Server *server;
-    caddr_t pack;
-    int packlen;
-    ZNotice_t *notice;
+server_forw_reliable(Server *server,
+		     caddr_t pack,
+		     int packlen,
+		     ZNotice_t *notice)
 {
     Code_t retval;
     Unacked *nacked;
+    int hashval;
 
     retval = ZSetDestAddr(&server->addr);
     if (retval != ZERR_NONE) {
@@ -1426,9 +1274,9 @@ server_forw_reliable(server, pack, packlen, notice)
 	syslog(LOG_WARNING, "srv_fwd xmit: %s", error_message(retval));
 	free(pack);
 	return;
-    }			
+    }
     /* now we've sent it, mark it as not ack'ed */
-		
+
     nacked = (Unacked *) malloc(sizeof(Unacked));
     if (!nacked) {
 	/* no space: just punt */
@@ -1437,13 +1285,15 @@ server_forw_reliable(server, pack, packlen, notice)
 	return;
     }
 
+    nacked->client = NULL;
     nacked->rexmits = 0;
     nacked->packet = pack;
     nacked->dest.srv_idx = server - otherservers;
     nacked->packsz = packlen;
     nacked->uid = notice->z_uid;
     nacked->timer = timer_set_rel(rexmit_times[0], srv_rexmit, nacked);
-    LIST_INSERT(srv_nacklist, nacked);
+    hashval = srv_nacktab_hashval(nacked->dest.srv_idx, nacked->uid);
+    Unacked_insert(&srv_nacktab[hashval], nacked);
 }
 
 /*
@@ -1451,8 +1301,7 @@ server_forw_reliable(server, pack, packlen, notice)
  */
 
 void
-server_send_queue(server)
-    Server *server;
+server_send_queue(Server *server)
 {
     Pending *pending;
     ZNotice_t notice;
@@ -1479,31 +1328,28 @@ server_send_queue(server)
  */
 
 static void
-srv_nack_cancel(notice, who)
-    ZNotice_t *notice;
-    struct sockaddr_in *who;
+srv_nack_cancel(ZNotice_t *notice,
+		struct sockaddr_in *who)
 {
-    Server *which = server_which_server(who);
+    Server *server = server_which_server(who);
     Unacked *nacked;
+    int hashval;
 
-    if (!which) {
+    if (!server) {
 	syslog(LOG_ERR, "non-server ack?");
 	return;
     }
-    for (nacked = srv_nacklist; nacked; nacked = nacked->next) {
-	if (&otherservers[nacked->dest.srv_idx] == which) {
-	    if (ZCompareUID(&nacked->uid, &notice->z_uid)) {
-		timer_reset(nacked->timer);
-		free(nacked->packet);
-		LIST_DELETE(nacked);
-		free(nacked);
-		return;
-	    }
+    hashval = srv_nacktab_hashval(server - otherservers, notice->z_uid);
+    for (nacked = srv_nacktab[hashval]; nacked; nacked = nacked->next) {
+	if (nacked->dest.srv_idx == server - otherservers
+	    && ZCompareUID(&nacked->uid, &notice->z_uid)) {
+	    timer_reset(nacked->timer);
+	    free(nacked->packet);
+	    Unacked_delete(nacked);
+	    free(nacked);
+	    return;
 	}
     }
-#if 0
-    zdbug((LOG_DEBUG, "srv_nack not found"));
-#endif
 }
 
 /*
@@ -1511,23 +1357,14 @@ srv_nack_cancel(notice, who)
  */
 
 static void
-srv_rexmit(arg)
-    void *arg;
+srv_rexmit(void *arg)
 {
     Unacked *packet = (Unacked *) arg;
     Code_t retval;
     /* retransmit the packet */
-	
-#if 0
-    zdbug((LOG_DEBUG,"srv_rexmit to %s/%d",
-	   otherservers[packet->dest.srv_idx].addr_str,
-	   ntohs(otherservers[packet->dest.srv_idx].addr.sin_port)));
-#endif
+
     if (otherservers[packet->dest.srv_idx].state == SERV_DEAD) {
-#if 0
-	zdbug((LOG_DEBUG, "cancelling send to dead server"));
-#endif
-	LIST_DELETE(packet);
+	Unacked_delete(packet);
 	free(packet->packet);
 	srv_nack_release(&otherservers[packet->dest.srv_idx]);
 	free(packet);
@@ -1556,22 +1393,20 @@ srv_rexmit(arg)
  */
 
 static void
-srv_nack_release(server)
-    Server *server;
+srv_nack_release(Server *server)
 {
-    /* XXX release any private queue for this server */
-
+    int i;
     Unacked *nacked, *next;
 
-    /* search the not-yet-acked list for anything destined to him, and
-       flush it. */
-    for (nacked = srv_nacklist; nacked; nacked = next) {
-	next = nacked->next;
-	if (&otherservers[nacked->dest.srv_idx] == server) {
-	    timer_reset(nacked->timer);
-	    free(nacked->packet);
-	    LIST_DELETE(nacked);
-	    free(nacked);
+    for (i = 0; i < SRV_NACKTAB_HASHSIZE; i++) {
+	for (nacked = srv_nacktab[i]; nacked; nacked = next) {
+	    next = nacked->next;
+	    if (nacked->dest.srv_idx == server - otherservers) {
+		timer_reset(nacked->timer);
+		Unacked_delete(nacked);
+		free(nacked->packet);
+		free(nacked);
+	    }
 	}
     }
 }
@@ -1582,23 +1417,24 @@ srv_nack_release(server)
  */
 
 static void
-srv_nack_renumber (new_idx)
-    int *new_idx;
+srv_nack_renumber (int *new_idx)
 {
     /* XXX release any private queue for this server */
     Unacked *nacked;
-    int idx;
+    int idx, i;
 
     /* search the not-yet-acked list for anything destined to 'from', and
        change the index to 'to'. */
-    for (nacked = nacklist; nacked; nacked = nacked->next) {
-	idx = new_idx[nacked->dest.srv_idx];
-	if (idx < 0) {
-	    syslog (LOG_ERR, "srv_nack_renumber error: [%d]=%d",
-		    nacked->dest.srv_idx, idx);
-	    idx = 0;
+    for (i = 0; i < SRV_NACKTAB_HASHSIZE; i++) {
+	for (nacked = srv_nacktab[i]; nacked; nacked = nacked->next) {
+	    idx = new_idx[nacked->dest.srv_idx];
+	    if (idx < 0) {
+		syslog(LOG_ERR, "srv_nack_renumber error: [%d]=%d",
+		       nacked->dest.srv_idx, idx);
+		idx = 0;
+	    }
+	    nacked->dest.srv_idx = idx;
 	}
-	nacked->dest.srv_idx = idx;
     }
 }
 
@@ -1606,12 +1442,11 @@ srv_nack_renumber (new_idx)
  * Queue this notice to be transmitted to the server when it is ready.
  */
 static void
-server_queue(server, len, pack, auth, who)
-    Server *server;
-    int len;
-    void *pack;
-    int auth;
-    struct sockaddr_in *who;
+server_queue(Server *server,
+	     int len,
+	     void *pack,
+	     int auth,
+	     struct sockaddr_in *who)
 {
     Pending *pending;
 
@@ -1638,11 +1473,10 @@ server_queue(server, len, pack, auth, who)
  */
 
 Pending *
-server_dequeue(server)
-    Server *server;
+server_dequeue(Server *server)
 {
     Pending *pending;
-	
+
     if (!server->queue)
 	return NULL;
     pending = server->queue;
@@ -1655,8 +1489,7 @@ server_dequeue(server)
  */
 
 void
-server_pending_free(pending)
-    Pending *pending;
+server_pending_free(Pending *pending)
 {
     free(pending->packet);
     free(pending);
@@ -1668,10 +1501,9 @@ server_pending_free(pending)
  */
 
 void
-server_self_queue(notice, auth, who)
-    ZNotice_t* notice;
-    int auth;
-    struct sockaddr_in * who;
+server_self_queue(ZNotice_t* notice,
+		  int auth,
+		  struct sockaddr_in * who)
 {
     char *pack;
     int packlen;
@@ -1691,8 +1523,7 @@ server_self_queue(notice, auth, who)
  * (true if called from signal handler)
  */
 void
-server_dump_servers(fp)
-    FILE *fp;
+server_dump_servers(FILE *fp)
 {
     int i;
 
@@ -1702,4 +1533,3 @@ server_dump_servers(fp)
 		otherservers[i].dumping ? " (DUMPING)" : "");
     }
 }
-
